{-# LANGUAGE CPP #-}
{-# LANGUAGE EmptyCase #-}
{-# LANGUAGE GADTs #-}
{-# LANGUAGE TemplateHaskell #-}
{-# OPTIONS_GHC -Wno-orphans #-}
{-# OPTIONS_GHC -Wno-unused-imports #-}
-- {-# OPTIONS_GHC -ddump-splices -dsuppress-all #-}
module Staged.SOP.Instances where

import Staged.SOP.Class
import Staged.SOP.TH
import Staged.Commons
import Generics.SOP

import Control.Exception
import Data.Char
import Data.Complex
import Data.Data
import Data.Fixed
import Data.Functor.Compose
import qualified Data.Functor.Const
import Data.Functor.Identity
import Data.Functor.Product
import Data.Functor.Sum
import Data.List.NonEmpty
import qualified Data.Monoid
import Data.Ord
import qualified Data.Semigroup
import Data.Version
import Data.Void
import Foreign.C.Error
import Foreign.C.Types
#if MIN_VERSION_base(4,11,0)
import GHC.ByteOrder
#endif
import GHC.Conc
import GHC.ExecutionStack
import GHC.Exts
-- import GHC.Events -- platform-specific, omitted
import GHC.Fingerprint
import GHC.Float
import qualified GHC.Generics
import GHC.IO.Buffer
import GHC.IO.Device
import GHC.IO.Encoding
import GHC.IO.Encoding.Failure
import GHC.IO.Exception
import GHC.IO.Handle
import GHC.RTS.Flags
import qualified GHC.Stack
import GHC.StaticPtr
import GHC.Stats
import System.Console.GetOpt
import System.IO
import Text.Printf
import Text.Read.Lex

-- Types from Generics.SOP:

deriveStagedGeneric ''I
deriveStagedGeneric ''K
deriveStagedGeneric ''(:.:)
deriveStagedGeneric ''(-.->)

-- Cannot derive instances for Sing
-- Cannot derive instances for Shape
-- Cannot derive instances for NP, NS, POP, SOP
-- Cannot derive instances for metadata types

-- Types from the Prelude:

deriveStagedGeneric ''Bool
deriveStagedGeneric ''Ordering
deriveStagedGeneric ''Maybe
deriveStagedGeneric ''Either
deriveStagedGeneric ''()
deriveStagedGeneric ''(,)              -- 2
deriveStagedGeneric ''(,,)
deriveStagedGeneric ''(,,,)
deriveStagedGeneric ''(,,,,)           -- 5
deriveStagedGeneric ''(,,,,,)
deriveStagedGeneric ''(,,,,,,)
deriveStagedGeneric ''(,,,,,,,)
deriveStagedGeneric ''(,,,,,,,,)
deriveStagedGeneric ''(,,,,,,,,,)      -- 10
deriveStagedGeneric ''(,,,,,,,,,,)
deriveStagedGeneric ''(,,,,,,,,,,,)
deriveStagedGeneric ''(,,,,,,,,,,,,)
deriveStagedGeneric ''(,,,,,,,,,,,,,)
deriveStagedGeneric ''(,,,,,,,,,,,,,,) -- 15
deriveStagedGeneric ''(,,,,,,,,,,,,,,,)
deriveStagedGeneric ''(,,,,,,,,,,,,,,,,)
deriveStagedGeneric ''(,,,,,,,,,,,,,,,,,)
deriveStagedGeneric ''(,,,,,,,,,,,,,,,,,,)
deriveStagedGeneric ''(,,,,,,,,,,,,,,,,,,,) -- 20
deriveStagedGeneric ''(,,,,,,,,,,,,,,,,,,,,)
deriveStagedGeneric ''(,,,,,,,,,,,,,,,,,,,,,)
deriveStagedGeneric ''(,,,,,,,,,,,,,,,,,,,,,,)
deriveStagedGeneric ''(,,,,,,,,,,,,,,,,,,,,,,,)
deriveStagedGeneric ''(,,,,,,,,,,,,,,,,,,,,,,,,) -- 25
deriveStagedGeneric ''(,,,,,,,,,,,,,,,,,,,,,,,,,)
deriveStagedGeneric ''(,,,,,,,,,,,,,,,,,,,,,,,,,,)
deriveStagedGeneric ''(,,,,,,,,,,,,,,,,,,,,,,,,,,,)
deriveStagedGeneric ''(,,,,,,,,,,,,,,,,,,,,,,,,,,,,)
deriveStagedGeneric ''(,,,,,,,,,,,,,,,,,,,,,,,,,,,,,) -- 30
deriveStagedGeneric ''[]

-- Other types from base:

-- From Control.Exception:
deriveStagedGeneric ''IOException
deriveStagedGeneric ''ArithException
deriveStagedGeneric ''ArrayException
deriveStagedGeneric ''AssertionFailed
deriveStagedGeneric ''AsyncException
deriveStagedGeneric ''NonTermination
deriveStagedGeneric ''NestedAtomically
deriveStagedGeneric ''BlockedIndefinitelyOnMVar
deriveStagedGeneric ''BlockedIndefinitelyOnSTM
deriveStagedGeneric ''AllocationLimitExceeded
deriveStagedGeneric ''Deadlock
deriveStagedGeneric ''NoMethodError
deriveStagedGeneric ''PatternMatchFail
deriveStagedGeneric ''RecConError
deriveStagedGeneric ''RecSelError
deriveStagedGeneric ''RecUpdError
deriveStagedGeneric ''ErrorCall
deriveStagedGeneric ''TypeError
deriveStagedGeneric ''MaskingState

-- From Data.Char:
deriveStagedGeneric ''GeneralCategory

-- From Data.Complex:
deriveStagedGeneric ''Complex

-- From Data.Data:
deriveStagedGeneric ''DataRep
deriveStagedGeneric ''Data.Data.Fixity
deriveStagedGeneric ''ConstrRep

-- From Data.Fixed:
deriveStagedGeneric ''Fixed
deriveStagedGeneric ''E0
deriveStagedGeneric ''E1
deriveStagedGeneric ''E2
deriveStagedGeneric ''E3
deriveStagedGeneric ''E6
deriveStagedGeneric ''E9
deriveStagedGeneric ''E12

-- From Data.Functor.Compose
deriveStagedGeneric ''Data.Functor.Compose.Compose

-- From Data.Functor.Const
deriveStagedGeneric ''Data.Functor.Const.Const

-- From Data.Functor.Identity
deriveStagedGeneric ''Identity

-- From Data.Functor.Product
deriveStagedGeneric ''Product

-- From Data.Functor.Sum
deriveStagedGeneric ''Sum

-- From Data.List.NonEmpty
deriveStagedGeneric ''NonEmpty

-- From Data.Monoid:
deriveStagedGeneric ''Data.Monoid.Dual
deriveStagedGeneric ''Data.Monoid.Endo
deriveStagedGeneric ''Data.Monoid.All
deriveStagedGeneric ''Data.Monoid.Any
deriveStagedGeneric ''Data.Monoid.Sum
deriveStagedGeneric ''Data.Monoid.Product
deriveStagedGeneric ''Data.Monoid.First
deriveStagedGeneric ''Data.Monoid.Last
deriveStagedGeneric ''Data.Monoid.Alt

-- From Data.Ord:
deriveStagedGeneric ''Down

-- From Data.Proxy:
deriveStagedGeneric ''Proxy

-- From Data.Semigroup:
deriveStagedGeneric ''Data.Semigroup.Min
deriveStagedGeneric ''Data.Semigroup.Max
deriveStagedGeneric ''Data.Semigroup.First
deriveStagedGeneric ''Data.Semigroup.Last
deriveStagedGeneric ''Data.Semigroup.WrappedMonoid
deriveStagedGeneric ''Data.Semigroup.Option
deriveStagedGeneric ''Data.Semigroup.Arg

-- From Data.Version:
deriveStagedGeneric ''Version

-- From Data.Void:
deriveStagedGeneric ''Void

-- From Foreign.C.Error:
deriveStagedGeneric ''Errno

-- From Foreign.C.Types:
deriveStagedGeneric ''CChar
deriveStagedGeneric ''CSChar
deriveStagedGeneric ''CUChar
deriveStagedGeneric ''CShort
deriveStagedGeneric ''CUShort
deriveStagedGeneric ''CInt
deriveStagedGeneric ''CUInt
deriveStagedGeneric ''CLong
deriveStagedGeneric ''CULong
deriveStagedGeneric ''CPtrdiff
deriveStagedGeneric ''CSize
deriveStagedGeneric ''CWchar
deriveStagedGeneric ''CSigAtomic
deriveStagedGeneric ''CLLong
deriveStagedGeneric ''CULLong
deriveStagedGeneric ''CIntPtr
deriveStagedGeneric ''CUIntPtr
deriveStagedGeneric ''CIntMax
deriveStagedGeneric ''CUIntMax
deriveStagedGeneric ''CClock
deriveStagedGeneric ''CTime
deriveStagedGeneric ''CUSeconds
deriveStagedGeneric ''CSUSeconds
deriveStagedGeneric ''CFloat
deriveStagedGeneric ''CDouble

#if MIN_VERSION_base(4,11,0)
-- From GHC.ByteOrder:
deriveStagedGeneric ''ByteOrder
#endif

-- From GHC.Conc:
deriveStagedGeneric ''ThreadStatus
deriveStagedGeneric ''BlockReason

-- From GHC.ExecutionStack:
deriveStagedGeneric ''Location
deriveStagedGeneric ''SrcLoc

-- From GHC.Exts:
deriveStagedGeneric ''RuntimeRep
deriveStagedGeneric ''VecCount
deriveStagedGeneric ''VecElem
deriveStagedGeneric ''SpecConstrAnnotation

-- From GHC.Generics:
deriveStagedGeneric ''GHC.Generics.K1
deriveStagedGeneric ''GHC.Generics.U1
deriveStagedGeneric ''GHC.Generics.V1
deriveStagedGeneric ''GHC.Generics.Par1
deriveStagedGeneric ''GHC.Generics.M1
deriveStagedGeneric ''GHC.Generics.R
deriveStagedGeneric ''GHC.Generics.S
deriveStagedGeneric ''GHC.Generics.D
deriveStagedGeneric ''GHC.Generics.C
deriveStagedGeneric ''(GHC.Generics.:*:)
deriveStagedGeneric ''(GHC.Generics.:+:)
deriveStagedGeneric ''(GHC.Generics.:.:)
deriveStagedGeneric ''GHC.Generics.Associativity
deriveStagedGeneric ''GHC.Generics.DecidedStrictness
deriveStagedGeneric ''GHC.Generics.SourceStrictness
deriveStagedGeneric ''GHC.Generics.SourceUnpackedness
deriveStagedGeneric ''GHC.Generics.Fixity

-- From GHC.IO.Buffer:
deriveStagedGeneric ''Buffer
deriveStagedGeneric ''BufferState

-- From GHC.IO.Device:
deriveStagedGeneric ''IODeviceType

-- From GHC.IO.Encoding:
deriveStagedGeneric ''BufferCodec
deriveStagedGeneric ''CodingProgress

-- From GHC.IO.Encoding.Failure:
deriveStagedGeneric ''CodingFailureMode

-- From GHC.Fingerprint
deriveStagedGeneric ''Fingerprint

-- From GHC.Float
deriveStagedGeneric ''FFFormat

-- From GHC.IO.Exception:
#if MIN_VERSION_base(4,11,0)
deriveStagedGeneric ''FixIOException
deriveStagedGeneric ''IOErrorType
#endif

-- From GHC.IO.Handle:
deriveStagedGeneric ''HandlePosn
#if MIN_VERSION_base(4,10,0)
deriveStagedGeneric ''LockMode
#endif

-- From GHC.RTS.Flags:
deriveStagedGeneric ''RTSFlags
deriveStagedGeneric ''GiveGCStats
deriveStagedGeneric ''GCFlags
deriveStagedGeneric ''ConcFlags
deriveStagedGeneric ''MiscFlags
deriveStagedGeneric ''DebugFlags
deriveStagedGeneric ''DoCostCentres
deriveStagedGeneric ''CCFlags
deriveStagedGeneric ''DoHeapProfile
deriveStagedGeneric ''ProfFlags
deriveStagedGeneric ''DoTrace
deriveStagedGeneric ''TraceFlags
deriveStagedGeneric ''TickyFlags
#if MIN_VERSION_base(4,10,0)
deriveStagedGeneric ''ParFlags
#endif

-- From GHC.Stack:
deriveStagedGeneric ''GHC.Stack.SrcLoc
deriveStagedGeneric ''GHC.Stack.CallStack

-- From GHC.StaticPtr:
deriveStagedGeneric ''StaticPtrInfo

-- From GHC.Stats:
#if MIN_VERSION_base(4,10,0)
deriveStagedGeneric ''RTSStats
deriveStagedGeneric ''GCDetails
#endif
#if !MIN_VERSION_base(4,11,0)
deriveStagedGeneric ''GCStats
#endif

-- From System.Console.GetOpt:

deriveStagedGeneric ''ArgOrder
deriveStagedGeneric ''OptDescr
deriveStagedGeneric ''ArgDescr

-- From System.Exit:

deriveStagedGeneric ''ExitCode

-- From System.IO:

deriveStagedGeneric ''IOMode
deriveStagedGeneric ''BufferMode
deriveStagedGeneric ''SeekMode
deriveStagedGeneric ''Newline
deriveStagedGeneric ''NewlineMode

-- From Text.Printf:

deriveStagedGeneric ''FieldFormat
deriveStagedGeneric ''FormatAdjustment
deriveStagedGeneric ''FormatSign
deriveStagedGeneric ''FormatParse

-- From Text.Read.Lex:

deriveStagedGeneric ''Lexeme
deriveStagedGeneric ''Number
